@@ -11,7 +11,7 @@ from scipy.spatial import distance
 from simtk.openmm import unit
 
 from .ommtk import MetadynamicSimulation
-from .utils import calc_coverage
+from .utils import calc_coverage, moving_average
 
 
 class SegmentMappingSim(MetadynamicSimulation):
@@ -45,7 +45,7 @@ class SegmentMappingSim(MetadynamicSimulation):
                  num_replicas=1, CV2_target_range = None,
                  time_per_iteration =10 * unit.nanosecond,
                  run_scans=True, run_production=True, input_scan_times=None, input_segment_coords=None,
-                 max_iterations=15, coverage_criteria=.5, min_coverage_criteria=.25, convergence_criteria=.1,
+                 max_iterations=15, max_coverage_criteria=.5, convergence_criteria=.1,
                  retain_scan_bias=True,
                  **kwargs):
 
@@ -71,8 +71,7 @@ class SegmentMappingSim(MetadynamicSimulation):
         self.CV2_target_range = CV2_target_range
         self.max_iterations = max_iterations
         self.retain_scan_bias = retain_scan_bias
-        self.coverage_criteria = coverage_criteria
-        self.min_coverage_criteria = min_coverage_criteria
+        self.max_coverage_criteria = max_coverage_criteria
         self.convergence_criteria = convergence_criteria
 
 
@@ -251,7 +250,7 @@ class SegmentMappingSim(MetadynamicSimulation):
         # for i in range(self.num_production_iterations):
         num_iterations = 0
         coverage = 0
-        while coverage<=self.coverage_criteria and num_iterations<=self.max_iterations:
+        while coverage<=self.max_coverage_criteria and num_iterations<=self.max_iterations:
             #get coverage first so that there's always one more iteration after coverage is reached
             coverage = calc_coverage(self.metaD.getFreeEnergy().value_in_unit(unit.kilocalorie_per_mole))
             print('coverage is currently {:.2f}%'.format(coverage * 100))
@@ -332,10 +331,15 @@ class SegmentMappingSim(MetadynamicSimulation):
 class SegmentMappingSimHREMD(SegmentMappingSim):
 
     def __init__(self, max_production_time=200 * unit.nanosecond, swap_freq = 2 * unit.nanosecond,
+                 convergence_window=2 * unit.nanosecond, dG_displacement_criteria = 1, min_coverage_criteria=.25,
                  **kwargs):
 
         self.max_production_time = max_production_time
         self.swap_freq = swap_freq
+        self.convergence_window = convergence_window
+        self.dG_displacement_criteria = dG_displacement_criteria
+        self.min_coverage_criteria = min_coverage_criteria
+
 
         super().__init__(**kwargs)
 
@@ -383,21 +387,9 @@ class SegmentMappingSimHREMD(SegmentMappingSim):
         attempted = 0
         continue_simulating=True
         dG_std = 100
+        dG_displacement = 100
         while continue_simulating:
 
-            #calc coverage, dG_std and detect convergence
-            coverage = calc_coverage(self.metaD.getFreeEnergy().value_in_unit(unit.kilocalorie_per_mole))
-            print('coverage is currently {:.2f}%'.format(coverage * 100))
-            with open(os.path.join(self.cwd,'segment_coords.pickle'),'rb') as handle:
-                dG_trace = pickle.load(handle)
-            #if fes_interval is .01 then 100 is variance across 2 ns
-            if len(dG_trace)>200:
-                dG_std = np.std(dG_trace[len(dG_trace-200):])
-
-            if coverage>self.min_coverage_criteria and dG_std < self.convergence_criteria:
-                print('convergence detected!')
-                continue_simulating= False
-
             #run the sim
             self.metaD.step(self.simulation, steps_per_swap_attempt)
             sim_time += self.swap_freq
@@ -429,6 +421,32 @@ class SegmentMappingSimHREMD(SegmentMappingSim):
                 self.simulation.context.setPositions(current_positions)
             print('HREMD acceptance rate {}'.format(accepted/attempted))
 
+            #calc coverage, dG_std and detect convergence
+            coverage = calc_coverage(self.metaD.getFreeEnergy().value_in_unit(unit.kilocalorie_per_mole))
+            print('coverage is currently {:.2f}%'.format(coverage * 100))
+            with open(os.path.join(self.cwd,'dG_trace.pickle'),'rb') as handle:
+                dG_trace = pickle.load(handle)
+
+            #get the rolling average dG displacement and variance of the production dG value trace over a limited range
+            # dG_trace is written every .01 * ns / step_length steps
+            window_steps = int(int(self.convergence_window / self.step_length)/int(.01 * unit.nanosecond/self.step_length))
+            # print('window steps, len(dG-trace)',window_steps,len(dG_trace))
+            if len(dG_trace)>window_steps:
+                dG_std = np.std(dG_trace[:-window_steps])
+                print('dG_std over the last {}: {}'.format(self.convergence_window,dG_std))
+                m_ave = moving_average(dG_trace[:-window_steps], np.max([1,int(len(dG_trace[:-window_steps])/10)]))
+                dG_displacement = np.max(m_ave) - np.min(m_ave)
+                print('dG_displacement over the last {}: {}'.format(self.convergence_window,dG_displacement))
+
+            #convergence detected when there is significant coverage and dG std and dG rolling average displacement is low
+            if coverage > self.min_coverage_criteria and \
+                dG_std < self.convergence_criteria and \
+                dG_displacement < self.dG_displacement_criteria:
+                print('convergence detected! Coverage: '
+                      '{}, dG_std: {}, dG_displacement: {}'.format(coverage,dG_std,dG_displacement))
+                continue_simulating= False
+
+            #or complete when max time is hit
             if sim_time>=self.max_production_time:
                 print('hit max sim time')
                 continue_simulating = False
